/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.hotmail.frojasg1.encrypting.encoderdecoder;

import com.frojasg1.general.GeneralException;
import com.frojasg1.general.HexadecimalFunctions;
import com.frojasg1.general.progress.CancellationException;
import com.frojasg1.general.progress.GeneralUpdatingProgress;
import com.frojasg1.general.progress.UpdatingProgress;
import com.frojasg1.general.progress.OperationCancellation;
import com.hotmail.frojasg1.encrypting.pseudorandomgenerator.PseudoRandomGenerator;
import com.hotmail.frojasg1.encrypting.pseudorandomgenerator.PseudoRandomGeneratorException;
import com.hotmail.frojasg1.encrypting.pseudorandomgenerator.chaoticgenerator.ChaoticGenerator_BigLong_depretated;

public class ReordererEncoder implements EncoderDecoder
{
	public ReordererEncoder() throws PseudoRandomGeneratorException
	{
		a_simpleEncoder = new SimpleEncoder();
		a_simpleReorderer = new SimpleReorderer();
	}
	
	public ReordererEncoder(int sizeOfNumbers, int numberOfBitsPerIteration ) throws PseudoRandomGeneratorException
	{
		a_simpleEncoder = new SimpleEncoder(sizeOfNumbers, numberOfBitsPerIteration);
		a_simpleReorderer = new SimpleReorderer(sizeOfNumbers, numberOfBitsPerIteration);
	}
	
	public ReordererEncoder(int sizeOfNumbersSimpleEncoder, int numberOfBitsPerIterationSimpleEncoder,
							int sizeOfNumbersReordererEncoder, int numberOfBitsPerIterationReordererEncoder )
		throws PseudoRandomGeneratorException
	{
		a_simpleEncoder = new SimpleEncoder(sizeOfNumbersSimpleEncoder, numberOfBitsPerIterationSimpleEncoder);
		a_simpleReorderer = new SimpleReorderer(sizeOfNumbersReordererEncoder, numberOfBitsPerIterationReordererEncoder);
	}

	// The parameter numberOfBitsPerIterationReordererEncoder must be the number of bits
	// returned by generatorReordererEncoder in every iteration of function M_nextPartOfByte
	public ReordererEncoder(	PseudoRandomGenerator generatorSimpleEncoder,
								PseudoRandomGenerator generatorReordererEncoder )
		 throws PseudoRandomGeneratorException
	{
		a_simpleEncoder = new SimpleEncoder( generatorSimpleEncoder );
		a_simpleReorderer = new SimpleReorderer( generatorReordererEncoder );
	}

	@Override
	public int M_getRecommendedNumberOfBytesToInitializeKey()
	{
		return( a_simpleReorderer.M_getRecommendedNumberOfBytesToInitializeKey() +
				a_simpleEncoder.M_getRecommendedNumberOfBytesToInitializeKey() );
	}

	@Override
	public byte[] getKey()
	{
		byte[] result = null;

		byte[] key1 = a_simpleEncoder.getKey();
		byte[] key2 = a_simpleEncoder.getKey();
		
		if( ( key1 != null ) && ( key2 != null ) )
			result = HexadecimalFunctions.M_joinByteArrays(key1, key2);

		return( result );
	}

	@Override
	public void M_initializeKey(byte[] newKey) throws PseudoRandomGeneratorException
	{
		int sizeKey1 = -1;
		int sizeKey2 = -1;
		
		if( newKey.length == M_getRecommendedNumberOfBytesToInitializeKey() )
		{
			sizeKey1 = a_simpleEncoder.M_getRecommendedNumberOfBytesToInitializeKey();
			sizeKey2 = a_simpleReorderer.M_getRecommendedNumberOfBytesToInitializeKey();
		}
		else
		{
			sizeKey1 = newKey.length/2;
			sizeKey2 = newKey.length-sizeKey1;
		}

		byte[] key1 = new byte[sizeKey1];
		byte[] key2 = new byte[sizeKey2];

		System.arraycopy( newKey, 0, key1, 0, sizeKey1 );
		System.arraycopy( newKey, sizeKey1, key2, 0, sizeKey2 );

		a_simpleEncoder.M_initializeKey(key1);
		a_simpleReorderer.M_initializeKey(key2);
	}

	@Override
	public void M_initializeKey(String newHexKey) throws PseudoRandomGeneratorException
	{
		try
		{
			byte[] newKey = HexadecimalFunctions.M_convertHexadecimalStringToByteArray( newHexKey );
			M_initializeKey(newKey);
		}
		catch( GeneralException ex )
		{
			ex.printStackTrace();
			throw( new PseudoRandomGeneratorException( ex.getMessage() ) );
		}
	}

	@Override
	public byte[] M_encode(byte[] input, UpdatingProgress up )
		throws PseudoRandomGeneratorException, CancellationException
	{
		GeneralUpdatingProgress gup = null;

		if( up != null )
		{
			gup = GeneralUpdatingProgress.new_GeneralUpdatingProgress();
			gup.up_setParentUpdatingProgress( up );
			gup.up_reset( 2.0D );
			gup.up_prepareNextSlice( 1.0D );
//			gup.setDebug(true);
		}

		byte[] inter = a_simpleEncoder.M_encode( input, gup );

		if( gup != null )
			gup.up_prepareNextSlice( 1.0D );

		byte[] result = a_simpleReorderer.M_encode( inter, gup );
		
		if( gup != null )
		{
			gup.up_performEnd();
			GeneralUpdatingProgress.store_GeneralUpdatingProgress( gup );
		}

		return( result );
	}

	@Override
	public byte[] M_decode(byte[] input, UpdatingProgress up )
		throws PseudoRandomGeneratorException, CancellationException
	{
		GeneralUpdatingProgress gup = null;
		
		if( up != null )
		{
			gup = GeneralUpdatingProgress.new_GeneralUpdatingProgress();
			gup.up_setParentUpdatingProgress( up );
			gup.up_reset( 2.0D );
			gup.up_prepareNextSlice( 1.0D );
//			gup.setDebug(true);
		}

		byte[] inter = a_simpleReorderer.M_decode( input, gup );

		if( gup != null )
			gup.up_prepareNextSlice( 1.0D );

		byte[] result = a_simpleEncoder.M_decode( inter, gup );

		if( gup != null )
		{
			gup.up_performEnd();
			GeneralUpdatingProgress.store_GeneralUpdatingProgress( gup );
		}

		return( result );
	}

	public SimpleEncoder getSimpleEncoder()
	{
		return( a_simpleEncoder );
	}

	public SimpleReorderer getSimpleReorderer()
	{
		return( a_simpleReorderer );
	}

	SimpleEncoder a_simpleEncoder;
	SimpleReorderer a_simpleReorderer;


	protected static void test1( String key, int sizeOfNumbers, int bitsPerIteration )
	{
		EncoderDecoder ed = null;
		
		try
		{
			PseudoRandomGenerator prg_se = null;
			PseudoRandomGenerator prg_sr = null;
				
			prg_se = new ChaoticGenerator_BigLong_depretated( sizeOfNumbers, bitsPerIteration );
			prg_sr = new ChaoticGenerator_BigLong_depretated( sizeOfNumbers, bitsPerIteration );
				
			ed = new ReordererEncoder( prg_se, prg_sr );

			ed.M_initializeKey( key );

			byte[] codedBytes = new byte[1000000];
			byte[] plainBytes = ed.M_decode( codedBytes, null );

			System.out.println( "Se acabo la decodificacion" );
		}
		catch( Throwable ex )
		{
			ex.printStackTrace();
		}
	}

	public static void main( String [] params )
	{
		test1(	"5ea7d62cc533aad00fe63825d0e6a488145857df6da8631dc46fd46534911403c04ef12e8dd98fd934fa3aa82bf5855a494f3cd0a9c6d1ac8cd6",
				4, 2 );

		try
		{
			ReordererEncoder re = new ReordererEncoder();
			String key = "La clave para reordenar mas larga de todas. a ver si funciona";
			byte[] newKey = key.getBytes();
			re.M_initializeKey(newKey);
			
			byte[] output = re.M_encode( newKey, null );
			System.out.println( "cadena codificada: " + new String( output ) );

			re.M_initializeKey(newKey);
			byte[] inputAgain = re.M_decode( output, null );
			System.out.println( "cadena decodificada again: " + new String( inputAgain ) );
		}
		catch( Throwable ex )
		{
			ex.printStackTrace();
		}
	}
}
