/* 
 * Copyright (C) 2017 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.hotmail.frojasg1.applications.fileencoder.fileencodertypes;

import com.frojasg1.general.GeneralException;
import com.frojasg1.general.HexadecimalFunctions;
import com.frojasg1.general.StringFunctions;
import com.frojasg1.general.progress.CancellationException;
import com.frojasg1.general.progress.GeneralUpdatingProgress;
import com.hotmail.frojasg1.applications.fileencoder.FileEncoderException;
import com.hotmail.frojasg1.applications.fileencoder.FileEncoderParameters;
import com.hotmail.frojasg1.encrypting.encoderdecoder.EncoderDecoder;
import com.hotmail.frojasg1.encrypting.encoderdecoder.ReordererEncoder;
import com.frojasg1.general.progress.UpdatingProgress;
import com.hotmail.frojasg1.applications.fileencoder.FileEncoder;
import com.hotmail.frojasg1.encrypting.pseudorandomgenerator.PseudoRandomGenerator;
import com.hotmail.frojasg1.encrypting.pseudorandomgenerator.PseudoRandomGeneratorException;
import com.hotmail.frojasg1.encrypting.pseudorandomgenerator.chaoticgenerator.ChaoticGenerator_BigLong;
import com.hotmail.frojasg1.encrypting.randomnumbers.RandomSource;
import java.io.File;
import java.util.Arrays;

/**
 *
 * @author axpe
 */
public class FileEncoderType_5 extends FileEncoderType_4
{
	protected long _decodedFileLength = -1;
	
	public FileEncoderType_5( UpdatingProgress up )
	{
		super( up );
	}

	public int M_getFileEncoderType()
	{
		return( FileEncoderParameters.INT_FILE_ENCODER_TYPE_5 );
	}

	protected byte[] M_encodeHeader_FET_5( byte[] encodedKey, byte[] signature,
									long fileLength ) throws FileEncoderException
	{
		byte[] header_head_buffer = a_fep.M_getHeaderHead().getBytes();
		
		byte[] variable_part_buffer = new byte[5];
		variable_part_buffer[0] = ( new Integer( a_fep.M_getFileEncoderType() ) ).byteValue();
		variable_part_buffer[1] = getByteValueOfSize( a_fep.M_getSizeOfNumbersSimpleEncoder() );
		variable_part_buffer[2] = getByteValueOfSize( a_fep.M_getSizeOfNumbersReordererEncoder() );
		variable_part_buffer[3] = ( new Integer( a_fep.M_getNumberOfBitsPerIterationSimpleEncoder() ) ).byteValue();
		variable_part_buffer[4] = ( new Integer( a_fep.M_getNumberOfBitsPerIterationReordererEncoder() ) ).byteValue();

		byte[] sliceSize_buffer = HexadecimalFunctions.M_getBufferFromInteger(a_fep.M_getNumberOfBytesOfFileSlice() );
		
		byte[] lengthOfFile_buffer = HexadecimalFunctions.M_getBufferFromLong( fileLength );
		
		byte[] result = new byte[ header_head_buffer.length + variable_part_buffer.length +
									sliceSize_buffer.length + lengthOfFile_buffer.length +
									encodedKey.length + signature.length ];

		System.arraycopy(header_head_buffer, 0, result, 0, header_head_buffer.length );
		System.arraycopy(variable_part_buffer, 0,
							result, header_head_buffer.length, variable_part_buffer.length );
		System.arraycopy(sliceSize_buffer, 0, result,
							header_head_buffer.length + variable_part_buffer.length,
							sliceSize_buffer.length );
		System.arraycopy(lengthOfFile_buffer, 0, result,
							header_head_buffer.length + variable_part_buffer.length +
							sliceSize_buffer.length,
							lengthOfFile_buffer.length );
		System.arraycopy(encodedKey, 0,
							result, header_head_buffer.length + variable_part_buffer.length +
							sliceSize_buffer.length + lengthOfFile_buffer.length,
							encodedKey.length );
		System.arraycopy( signature, 0, result,	header_head_buffer.length +
									variable_part_buffer.length +
									sliceSize_buffer.length + lengthOfFile_buffer.length +
									encodedKey.length,
									signature.length );

		return( result );
	}

	protected void M_updateParametersWithHeader( byte[] header, String fileName ) throws FileEncoderException
	{
		int lenHead = a_fep.M_getHeaderHead().getBytes().length;
		
		byte[] fileHead_buffer = new byte[lenHead];
		byte[] variable_part_buffer = new byte[5];
		byte[] sliceSize_buffer = new byte[4];
		byte[] lengthOfFile_buffer = new byte[8];

		System.arraycopy(header, 0, fileHead_buffer, 0, lenHead );
		System.arraycopy(header, lenHead, variable_part_buffer, 0, 5 );
		System.arraycopy(header, lenHead + variable_part_buffer.length, sliceSize_buffer, 0, 4 );
		System.arraycopy(header, lenHead + variable_part_buffer.length + sliceSize_buffer.length,
							lengthOfFile_buffer, 0, 8);

		String fileHead = new String( fileHead_buffer );
		if( ! fileHead.equals( a_fep.M_getHeaderHead() ) )	throw( new FileEncoderException( "HeaderHead unexpected in file " + fileName ) );
		
		int sliceSize = -1;
		try
		{
			sliceSize = HexadecimalFunctions.M_getIntegerFromBuffer(sliceSize_buffer);
		}
		catch( GeneralException ge )
		{
			ge.printStackTrace();
			throw( new FileEncoderException( "Error getting sliceSize from buffer in file: " + fileName ) );
		}

		try
		{
			_decodedFileLength = HexadecimalFunctions.M_getLongFromBuffer(lengthOfFile_buffer);
		}
		catch( GeneralException ge )
		{
			ge.printStackTrace();
			throw( new FileEncoderException( "Error getting size of file from buffer in file: " + fileName ) );
		}

		int fileEncoderType = variable_part_buffer[0];
		byte sizeOfNumbersSimpleEncoder = variable_part_buffer[1];
		byte sizeOfNumbersReordererEncoder = variable_part_buffer[2];
		int numberOfBitsPerIterationSimpleEncoder = variable_part_buffer[3];
		int numberOfBitsPerIterationReordererEncoder = variable_part_buffer[4];

		if( fileEncoderType != M_getFileEncoderType() )	throw( new FileEncoderException( "FileEncoderType unexpected in file " + fileName ) );

		a_fep.M_setFileEncoderType(fileEncoderType);
		a_fep.M_setSliceSize(sliceSize);
		a_fep.M_setSizeOfNumbersSimpleEncoder( getUnsignedValueForSize( sizeOfNumbersSimpleEncoder ) );
		a_fep.M_setSizeOfNumbersReordererEncoder( getUnsignedValueForSize( sizeOfNumbersReordererEncoder ) );
		a_fep.M_setNumberOfBitsPerIterationSimpleEncoder(numberOfBitsPerIterationSimpleEncoder);
		a_fep.M_setNumberOfBitsPerIterationReordererEncoder(numberOfBitsPerIterationReordererEncoder);
	}

	@Override
	public void M_encodeFile( char[] password, String fileName, String newFileName ) throws FileEncoderException, CancellationException
	{
		try
		{
			if( ( new File( newFileName )).exists() )	throw( new FileEncoderException( "ERROR!! File exists: "+newFileName ) );
			a_encoder = M_newEncoderDecoder(a_fep);
			int keySize = a_encoder.M_getRecommendedNumberOfBytesToInitializeKey();

			long fileLength = FileEncoder.getFileLength(fileName);
			GeneralUpdatingProgress gup = null;
			gup = new GeneralUpdatingProgress();
//			gup.up_setDebug( true );
			gup.up_setParentUpdatingProgress( a_updatingProgress );
			gup.up_reset( fileLength * 1.01D );
			gup.up_prepareNextSlice( fileLength * 0.01D );
			gup.up_setMinimumPercentageToReport( FileEncoder.getMinimumFractionToReport(fileLength) );

			byte[] key = RandomSource.instance().M_getRandomBytes(keySize);
//			a_encoder = M_newEncoderDecoder_forShortKey( a_fep.M_getPassword().getBytes().length );
//			a_encoder.M_initializeKey( HexadecimalFunctions.M_joinByteArrays(a_fep.M_getPassword().getBytes(), a_fep.M_getPassword().getBytes() ) );
			a_encoder = M_newEncoderDecoder_forShortKey( 32 ); // number of bytes returned by sha-256

			if( password == null )	password = a_fep.M_getPassword();
			if( password == null )	throw( new FileEncoderException( "Password needed" ) );
			
			a_encoder.M_initializeKey( M_calculateSHA_256( StringFunctions.M_getBytes(password) ) );

			byte[] encodedKey = a_encoder.M_encode( key, null );

			a_encoder = M_newEncoderDecoder(a_fep);
			a_encoder.M_initializeKey( key );

			int numberOfBytesFileSlice = a_fep.M_getNumberOfBytesOfFileSlice();

			byte[] md5 =  FileEncoder.M_getMD5_fromFile_sliced( fileName, numberOfBytesFileSlice, gup );
			byte[] encodedMD5 = a_encoder.M_encode( md5, null );

			byte[] header = M_encodeHeader_FET_5( encodedKey, encodedMD5, fileLength );

			gup.up_prepareNextSlice( fileLength );
			int[] randomOrder = getRandomOrder( numberOfBytesFileSlice, fileLength );
		    FileEncoder.M_writeEncodedFile_randomlySliced( fileName, newFileName, header, randomOrder,
															a_encoder, numberOfBytesFileSlice, gup );

			gup.up_performEnd();
		}
		catch( PseudoRandomGeneratorException ex )
		{
			ex.printStackTrace();
			throw( new FileEncoderException( ex.getMessage() ) );
		}
	}

	protected int[] getRandomOrder( int sliceSize, long fileLength ) throws FileEncoderException, PseudoRandomGeneratorException, CancellationException
	{
		int[] randomOrder = null;
		if( sliceSize > 0 )
		{
			int numberOfSlices = (int) ( ( fileLength - 1 ) / sliceSize + 1 );
			randomOrder = ( (ReordererEncoder) a_encoder).getSimpleReorderer().getRandomOrder( numberOfSlices, null);
		}
		return( randomOrder );
	}

	@Override
	public void M_decodeFile( char[] password, String fileName, String newFileName ) throws FileEncoderException, CancellationException
	{
		int offsetEncodedKey = getOffsetForEncodedKey();
		int keySize = 0;

		try
		{
			if( ( new File( newFileName )).exists() )	throw( new FileEncoderException( "ERROR!! File exists: "+newFileName ) );

			// we read the header, with the fixed amount of bytes of 1 KiB, assuring that the whole header will be read
			byte[] header = FileEncoder.M_getHeaderFromFile( fileName, 8192 );
			M_updateParametersWithHeader( header, a_fep.M_getEncodedFileName() );

			long fileLength = FileEncoder.getFileLength(fileName);
			GeneralUpdatingProgress gup = null;
			gup = new GeneralUpdatingProgress();
			gup.up_setParentUpdatingProgress( a_updatingProgress );
			gup.up_reset( fileLength * 1.01D );
			gup.up_prepareNextSlice( fileLength );
			gup.up_setMinimumPercentageToReport( FileEncoder.getMinimumFractionToReport(fileLength) );

			a_encoder = M_newEncoderDecoder(a_fep);
			keySize = a_encoder.M_getRecommendedNumberOfBytesToInitializeKey();

			if( header.length <= offsetEncodedKey + keySize + 16 )
			{
				throw( new FileEncoderException( "File too short. It is not a JavaFileEncoder file, FileName:"+ fileName ) );
			}
			
			byte[] encodedKey = new byte[ keySize ];
			System.arraycopy(header, offsetEncodedKey, encodedKey, 0, keySize );
			byte[] encodedMD5 = new byte[16];
			System.arraycopy(header, offsetEncodedKey + keySize, encodedMD5, 0, 16 );
			
//			a_encoder = M_newEncoderDecoder_forShortKey( a_fep.M_getPassword().getBytes().length );
//			a_encoder.M_initializeKey( HexadecimalFunctions.M_joinByteArrays(a_fep.M_getPassword().getBytes(), a_fep.M_getPassword().getBytes() ) );
			a_encoder = M_newEncoderDecoder_forShortKey( 32 ); // number of bytes returned by sha-256
			
			if( password == null )	password = a_fep.M_getPassword();
			if( password == null )	throw( new FileEncoderException( "Password needed" ) );
			
			a_encoder.M_initializeKey( M_calculateSHA_256( StringFunctions.M_getBytes(password) ) );

			byte[] key = a_encoder.M_decode( encodedKey, null );

//			System.out.println( HexadecimalFunctions.M_convertByteArrayToHexadecimalString( key ) );

			a_encoder = M_newEncoderDecoder(a_fep);
			a_encoder.M_initializeKey( key );

		    byte[] md5 =  a_encoder.M_decode( encodedMD5, null );

			int numberOfBytesFileSlice = a_fep.M_getNumberOfBytesOfFileSlice();

//			byte[] calculatedMD5 = new byte[16];

			int[] randomOrder = getRandomOrder( numberOfBytesFileSlice, _decodedFileLength );
			FileEncoder.M_writeDecodedFile_randomlySliced( fileName, newFileName,
													offsetEncodedKey + a_encoder.M_getRecommendedNumberOfBytesToInitializeKey() + 16,
													_decodedFileLength,
													randomOrder,
													a_encoder,
													numberOfBytesFileSlice, gup );

			gup.up_prepareNextSlice( fileLength * 0.01D );
			byte[] calculatedMD5 = FileEncoder.M_getMD5_fromFile_sliced( newFileName, numberOfBytesFileSlice, gup );
		    if( ! Arrays.equals(md5, calculatedMD5) )
			{
				long size = 0;
				try
				{
					File file = new File( newFileName );
					size = file.length();
					file.delete();
				}
				catch( Throwable th )
				{
					th.printStackTrace();
				}
		    	throw( new FileEncoderException( "MD5 signature not correct. Aborting. File: " + fileName + " with length: " + size ) );
			}

		    System.out.println( "MD5 signature correct !!  Recovered unencrypted file: " + newFileName );

			gup.up_performEnd();
		}
		catch( PseudoRandomGeneratorException ex )
		{
			ex.printStackTrace();
			throw( new FileEncoderException( ex.getMessage() ) );
		}
	}

	@Override
	public void M_encodeFileFromBuffer( char[] password, FileEncoderParameters fep,
										byte[] dataToEncode,
										String fileName ) throws FileEncoderException, CancellationException
	{
		try
		{
			a_fep = fep;

			a_encoder = M_newEncoderDecoder(a_fep);
			int keySize = a_encoder.M_getRecommendedNumberOfBytesToInitializeKey();

			byte[] key = RandomSource.instance().M_getRandomBytes(keySize);
			a_encoder = M_newEncoderDecoder_forShortKey( 32 ); // number of bytes returned by sha-256

			if( password == null )	password = a_fep.M_getPassword();
			if( password == null )	throw( new FileEncoderException( "Password needed" ) );

			a_encoder.M_initializeKey( M_calculateSHA_256( StringFunctions.M_getBytes(password) ) );

			byte[] encodedKey = a_encoder.M_encode( key, null );

			a_encoder = M_newEncoderDecoder(a_fep);
			a_encoder.M_initializeKey( key );

			int numberOfBytesFileSlice = a_fep.M_getNumberOfBytesOfFileSlice();

			byte[] md5 = null;

			md5 =  FileEncoder.M_getMD5( dataToEncode );

			byte[] encodedMD5 = a_encoder.M_encode( md5, null );

			byte[] header = M_encodeHeader_FET_5( encodedKey, encodedMD5, dataToEncode.length );

			int[] randomOrder = getRandomOrder( numberOfBytesFileSlice, dataToEncode.length );
			FileEncoder.encodeDataToFile_randomlySliced( fileName, header, a_encoder,
															numberOfBytesFileSlice,
															a_updatingProgress,
															dataToEncode,
															randomOrder );
		}
		catch( PseudoRandomGeneratorException ex )
		{
			ex.printStackTrace();
			throw( new FileEncoderException( ex.getMessage() ) );
		}
	}

	@Override
	public byte[] M_decodeFileToBuffer( char[] password, FileEncoderParameters fep,
										String fileName ) throws FileEncoderException, CancellationException
	{
		a_fep = fep;

		byte[] decodedData = null;

		int offsetEncodedKey = getOffsetForEncodedKey();
		int keySize = 0;

		try
		{
			// we read the header, with the fixed amount of bytes of 1 KiB, assuring that the whole header will be read
			byte[] header = FileEncoder.M_getHeaderFromFile( fileName, 8192 );
			M_updateParametersWithHeader( header, fileName );

			a_encoder = M_newEncoderDecoder(a_fep);
			keySize = a_encoder.M_getRecommendedNumberOfBytesToInitializeKey();

			if( header.length < offsetEncodedKey + keySize + 16 )
			{
				throw( new FileEncoderException( "File too short. It is not a JavaFileEncoder file, FileName:"+ fileName ) );
			}

			byte[] encodedKey = new byte[ keySize ];
			System.arraycopy(header, offsetEncodedKey, encodedKey, 0, keySize );
			byte[] encodedMD5 = new byte[16];
			System.arraycopy(header, offsetEncodedKey + keySize, encodedMD5, 0, 16 );
			
//			a_encoder = M_newEncoderDecoder_forShortKey( a_fep.M_getPassword().getBytes().length );
//			a_encoder.M_initializeKey( HexadecimalFunctions.M_joinByteArrays(a_fep.M_getPassword().getBytes(), a_fep.M_getPassword().getBytes() ) );
			a_encoder = M_newEncoderDecoder_forShortKey( 32 ); // number of bytes returned by sha-256
			
			if( password == null )	password = a_fep.M_getPassword();
			if( password == null )	throw( new FileEncoderException( "Password needed" ) );
			
			a_encoder.M_initializeKey( M_calculateSHA_256( StringFunctions.M_getBytes(password) ) );

			byte[] key = a_encoder.M_decode( encodedKey, null );

//			System.out.println( HexadecimalFunctions.M_convertByteArrayToHexadecimalString( key ) );

			a_encoder = M_newEncoderDecoder(a_fep);
			a_encoder.M_initializeKey( key );

		    byte[] md5 =  a_encoder.M_decode( encodedMD5, null );

			int numberOfBytesFileSlice = a_fep.M_getNumberOfBytesOfFileSlice();

			int[] randomOrder = getRandomOrder( numberOfBytesFileSlice, _decodedFileLength );
			decodedData = FileEncoder.decodeEncryptedFile_randomlySliced( fileName,
														offsetEncodedKey + a_encoder.M_getRecommendedNumberOfBytesToInitializeKey() + 16,
														(int) _decodedFileLength, randomOrder,
														a_encoder,
														numberOfBytesFileSlice, a_updatingProgress );

			if( decodedData != null )
			{
//				decodedData = a_encoder.M_decode( encodedData, a_updatingProgress, oc );

				byte[] calculatedMD5 = FileEncoder.M_getMD5( decodedData );

				if( ! Arrays.equals(md5, calculatedMD5) )
				{
	/*				try
					{
						File file = new File( newFileName );
						file.delete();
					}
					catch( Throwable th )
					{
						th.printStackTrace();
					}
	*/		    	throw( new FileEncoderException( "MD5 signature not correct. Aborting. File: " + fileName ) );
				}

				System.out.println( "MD5 signature correct !!  Recovered unencrypted file: " + fileName );
			}
		}
		catch( Throwable ex )
		{
			ex.printStackTrace();
			throw( new FileEncoderException( ex.getMessage() ) );
		}

		return( decodedData );
	}

	protected int getOffsetForEncodedKey()
	{
		return( FileEncoderParameters.M_getHeaderHead().getBytes().length + 9 + 8 );
	}


}
